/* GCQuery -- Gene Cluster query
	Input: 
	1. A file that specifies the query gene cluster. This file 
	   contains gene names and protein sequences in FASTA format.
	2. One or more FASTA files that contains all protein names 
	   and sequences on each chromosome. Proteins should be in 
	   the same order as on the chromosome.

	Return:  Significant gene clusters with e-values from low to high. 

	Reference: Yang Q. and Sze S.-H. (2008) Large-scale analysis of gene clustering in bacteria. Genome Research. 

	Author: Qingwu Yang
	Department of Computer Science
	Texas A&M University, Texas USA
	April, 2008
**/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include "readtxt.h"
#include "calHyperGeometryEval.h"

//#define GCQUERY_DEBUG

void infobox(char **v);

void cmd_parser(int n, char ** v, char * queryfile, char *** dbfiles,int *n_dbfiles, char * outputfile, int *chrForm, double *b, double *g);

int main(int argc, char ** argv){
#ifdef GCQUERY_DEBUG
	FILE * timeFile;
#endif

	/* Whether a chromosome is circular or not 
	 */
	int isCircularChromosome = 0;  

	//query gene cluster
	char gcfname[100];  //name of the file containing the query gene cluster (in fasta format, name + protein sequence)
	T_PROTEIN * gcproteins;
	int         n_gcproteins;

	//queried chromosomes
	char chrListFilename[100];
	char ** chrFileList; //filenames of all chromosome files
	int  n_gene;         //total # of genes on a genome
	int  n_hit;          //total # of genes associated with a query gene (#hits)
	int  n_chr;          //number of chromosomes
	int  * chr_sizes;    //size of each chr
	int  * maxwinsizes;     //max possible window size
	T_POSITION ** chr_hits;  //hits on each chr

	int i, s, e, n, max;
	T_POSITION * tmppos;

	double ** EVALUES;

	//results
	T_GCLUSTER ** gclusters; 
	char resultfilename[250];

#ifdef GCQUERY_DEBUG
	time_t time0, time1, time2, time3; 
	double diff1, diff2, diff3;

	time0 = time(NULL);
#endif

	strcpy(gcfname, "genecluster");
	strcpy(resultfilename, "geneclusters.txt");
	strcpy(chrListFilename, "chrfiles");

	cmd_parser(argc,argv, gcfname, &chrFileList, &n_chr, resultfilename, &isCircularChromosome, &E_VALUE_CUTOFF, &GENE_CLUSTER_E_VALUE_CUTOFF);
	
	//read in query gene cluster
	sequence_reader_fasta(gcfname, &gcproteins, &n_gcproteins);

	//get all hits first
	if( !(chr_sizes = calloc(n_chr, sizeof(int))) )
		fprintf(stderr, "\nout of memory for chr_sizes!\n\n"), fflush(NULL), exit(0);
	if( !(maxwinsizes = calloc(n_chr, sizeof(int))) )
		fprintf(stderr, "\nout of memory for maxwinsizes!\n\n"), fflush(NULL), exit(0);
	if( !(chr_hits = calloc(n_chr, sizeof(T_POSITION *))) )
		fprintf(stderr, "\nout of memory for chr_hits!\n\n"), fflush(NULL), exit(0);

#ifdef GCQUERY_DEBUG
	time1 = time(NULL);
#endif

	n_hit = n_gene = max = 0;
	for(i=0; i<n_chr; ++i){
		//process a chromosome in a .faa file and get a pointer to T_POSITION containing position information etc. for next step
		process_chromosome(chrFileList[i], gcproteins, n_gcproteins, chr_hits+i, chr_sizes+i);

		//cal total # of genes
		n_gene += chr_sizes[i]; 

		if(chr_hits[i] == NULL)
			continue;

		tmppos = chr_hits[i]; 
		s=e=tmppos->pos;
		n = 0;

		while( tmppos ){
			if(tmppos->pos > e)
				e = tmppos->pos;

			n++;
			tmppos = tmppos->next;
		}

		n_hit += n;

		maxwinsizes[i] = e - s +1;

		if( maxwinsizes[i] > max )
			max = maxwinsizes[i];
	}

#ifdef GCQUERY_DEBUG
	time2 = time(NULL);
#endif

	if( n_hit && n_gene ){
		//calculate the e-value table 
		comput_evalues(n_gene, n_hit, max, &EVALUES, isCircularChromosome);

		//find gene clusters on each chr
		if( !(gclusters = calloc(n_chr, sizeof(T_GCLUSTER *))) )
			fprintf(stderr, "\nout of memory for gclusters!\n\n"), fflush(NULL), exit(0);

		for(i=0; i<n_chr; ++i){
			//find gene cluster in a chromosome and get a pointer to T_RESULT containing resulting clusters
			if(isCircularChromosome == 1)
				find_clusters_inOneCircularChr(chr_hits[i], EVALUES, maxwinsizes[i], gclusters+i, chr_sizes[i]);
			else
				find_clusters_inonechr(chr_hits[i], EVALUES, maxwinsizes[i], gclusters+i);

			//remove repeats and subsets in resulting clusters
			remove_subclusters_inonechr(gclusters[i]);
		}

		//output resulting clusters
		output_clusters(gcproteins, n_gcproteins, gclusters, n_chr, resultfilename, chrFileList); 
	}

	remove("tmpfaafile_xxx_tobedeleted");
	remove("tmpfaafile_xxx_tobedeleted.phr");
	remove("tmpfaafile_xxx_tobedeleted.pin");
	remove("tmpfaafile_xxx_tobedeleted.psq");
	remove("formatdb.log");

#ifdef GCQUERY_DEBUG
	time3 = time(NULL);

	diff1 = difftime(time1, time0);
	diff2 = difftime(time2, time1);
	diff3 = difftime(time3, time2);

	timeFile = fopen("timetest", "a");
	fprintf(timeFile, "%s:\t%f\t%f\t%f\n", chrListFilename, diff1, diff2, diff3);
	fclose(timeFile);
#endif

	return 0;
}

void infobox(char ** v){
	fprintf(stdout, "\n\n%s ", v[0]);
	fprintf(stdout, "\n\t-q \"file name containing query cluster\" ");
	fprintf(stdout, "\n\t-d \"file names containing all gene names and protein sequences on each chromosome\" ");
	fprintf(stdout, "\n\t-c \"type of genome (0 - linear; 1 - circular)\" ");
	fprintf(stdout, "\n\t-b \"BLAST e-value cutoff\" ");
	fprintf(stdout, "\n\t-e \"GCQuery e-value cutoff\" ");
	fprintf(stdout, "\n\t-o \"output file name\" ");
	fprintf(stdout, "\n\n");

	fflush(NULL), exit(0); 
}


void cmd_parser(int n, char ** v, char * queryfile, char *** dbfiles, int *n_dbfiles, char * outputfile, int *chrForm, double *b, double *g){
	int i, j, k;

	if( n < 5 ){
		infobox(v);
	} else {
		for(i=1; i<n; ++i){
			if( strstr(v[i], "-q") ){
				strcpy(queryfile, v[i+1]);
				++i;
			} else if( strstr(v[i], "-c") ){
				*chrForm = atoi(v[i+1]);
				++i;
			} else if( strstr(v[i], "-d") ){
				for(k=0, j=i+1; j<n && v[j][0] != '-'; ++j){
					++k;
				}

				if( k == 0 )
					infobox(v);

				*n_dbfiles = k;

				if( !(*dbfiles = malloc(k * sizeof(char *))) )
					fprintf(stdout, "\n\nout of memory for dbfiles\n"), fflush(NULL),exit(0);

				for(k=0, j=i+1; j<n && v[j][0] != '-'; ++j){
					if( !((*dbfiles)[k] = malloc((strlen(v[j])+1)*sizeof(char))))
						fprintf(stdout, "\n\nout of memory for dbfiles[]\n"), fflush(NULL),exit(0);

					strcpy((*dbfiles)[k], v[j]);
					++k;
				}

				i = j-1;
			} else if( strstr(v[i], "-o") ){
				strcpy(outputfile, v[i+1]);
				++i;
			} else if( strstr(v[i], "-b") ){
				*b = atof(v[i+1]);
				++i;
			} else if( strstr(v[i], "-e") ){
				*g = atof(v[i+1]);
				++i;
			} else {
				infobox(v);
			}
		}
	}

	return; 	
}


